/* 
	ctrlproxy: A modular IRC proxy
	(c) 2002-2003 Jelmer Vernooij <jelmer@nl.linux.org>

	This program is free software; you can redistribute it and/or modify
	it under the terms of the GNU General Public License as published by
	the Free Software Foundation; either version 2 of the License, or
	(at your option) any later version.

	This program is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with this program; if not, write to the Free Software
	Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*/

#include "internals.h"
#include "config.h"

#include <libxml/xmlmemory.h>
#include <libxml/parser.h>

#define add_log_domain(domain) g_log_set_handler (domain, G_LOG_LEVEL_MASK | G_LOG_FLAG_FATAL | G_LOG_FLAG_RECURSION, log_handler, NULL);

/* globals */
char my_hostname[MAXHOSTNAMELEN+2];
xmlNodePtr xmlNode_networks, xmlNode_plugins;
GHookList data_hook;
xmlDocPtr configuration;
char *configuration_file;
FILE *debugfd = NULL;
FILE *f_logfile = NULL;

GList *plugins = NULL;

void signal_crash(int sig) 
{
	g_log(G_LOG_DOMAIN, G_LOG_LEVEL_ERROR, "Received SIGSEGV!\n");
#ifdef DEBUG
	g_log(G_LOG_DOMAIN, G_LOG_LEVEL_ERROR, "SIGSEGV in ctrlproxy core, waiting for gdb connection...\n");
	/* Bug in the core. Just loop until someone runs gdb on us */ 
	while(1)sleep(5);
#else
	g_log(G_LOG_DOMAIN, G_LOG_LEVEL_CRITICAL, "SIGSEGV in ctrlproxy core, exiting\n");
	exit(1);
#endif
}

void log_handler(const gchar *log_domain, GLogLevelFlags flags, const gchar *message, gpointer user_data) {
	if(!f_logfile)return;
	fprintf(f_logfile, "%s\n", message);
	fflush(f_logfile);
}

void signal_quit(int sig)
{
	g_log(G_LOG_DOMAIN, G_LOG_LEVEL_MESSAGE, "Received signal %d, quitting...\n", sig);
	if(debugfd)fclose(debugfd);
	exit(0);
}

void unload_plugin(struct plugin *p)
{
	plugin_fini_function f;

	/* Run exit function if present */
	if(g_module_symbol(p->module, "fini_plugin", (gpointer *)&f)) {
		f(p->module, p->xmlConf);
	}

	g_module_close(p->module);

	/* Remove autoload from this plugins' element */
	xmlUnsetProp(p->xmlConf, "autoload");
}

void load_plugin(xmlNodePtr cur)
{
	GModule *m;
	char *mod_name;
	struct plugin *p;
	char *modulesdir;
	gchar *path_name;
	plugin_init_function f; 

	mod_name = xmlGetProp(cur, "file");
	if(!mod_name) {
		g_warning("No filename specified for plugin");
		return;
	}

	/* Determine correct modules directory */
	if(getenv("MODULESDIR"))modulesdir = getenv("MODULESDIR"); 
	else modulesdir = MODULESDIR;

	if(mod_name[0] == '/')path_name = g_strdup(mod_name);
	else path_name = g_module_build_path(modulesdir, mod_name);


	m = g_module_open(path_name, 0);
	if(!m) {
		g_warning("Unable to open module %s(%s), ignoring\n", path_name, g_module_error()); 
		xmlFree(mod_name);
		g_free(path_name);
		return;
	} else {
		if(!g_module_symbol(m, "init_plugin", (gpointer *)&f)) {
			g_warning("Can't find symbol 'init_plugin' in module %s\n", path_name);
		} else f(m, cur);
	}

	g_free(path_name);

	p = malloc(sizeof(struct plugin));
	p->name = strdup(mod_name);
	p->module = m;
	p->xmlConf = cur;

	plugins = g_list_append(plugins, p);
	
	xmlSetProp(cur, "autoload", "1");
	xmlFree(mod_name);
}

void save_configuration()
{
	xmlSaveFile(configuration_file, configuration);
}

void signal_save(int sig)
{
	save_configuration();
}

void readConfig(char *file) {
    xmlNodePtr cur;

	configuration = xmlParseFile(file);
	g_assert(configuration);

	cur = xmlDocGetRootElement(configuration);
	g_assert(cur);

	g_assert(!strcmp(cur->name, "ctrlproxy"));

	cur = cur->xmlChildrenNode;
	while(cur) {
		if(xmlIsBlankNode(cur) || !strcmp(cur->name, "comment")) {
			cur = cur->next;
			continue;
		}
		
		if(!strcmp(cur->name, "plugins")) {
			xmlNode_plugins = cur;
		} else if(!strcmp(cur->name, "networks")) {
			xmlNode_networks = cur;
		} else g_assert(0);
		
		cur = cur->next;
	}
}

int main(int argc, const char *argv[])
{
	GMainLoop *loop;
	char daemon = 0;
	char *logfile = NULL, *rcfile = NULL;
#ifdef HAVE_POPT_H
	int c;
	poptContext pc;
	struct poptOption options[] = {
		POPT_AUTOHELP
		{"debug", 'd', POPT_ARG_STRING, NULL, 'd', "Write irc traffic to specified file", "FILE" },
		{"daemon", 'D', POPT_ARG_VAL, &daemon, 1, "Run in the background (as a daemon)"},
		{"log", 'l', POPT_ARG_STRING, &logfile, 0, "Log messages to specified file", "FILE"},
		{"rc-file", 'r', POPT_ARG_STRING, &rcfile, 0, "Use configuration file from specified location", "FILE"},
		{"version", 'v', POPT_ARG_NONE, NULL, 'v', "Show version information"},
		POPT_TABLEEND
	};
#endif
    xmlNodePtr cur;

	signal(SIGINT, signal_quit);
	signal(SIGTERM, signal_quit);
	signal(SIGPIPE, SIG_IGN);
	signal(SIGHUP, SIG_IGN);
	signal(SIGSEGV, signal_crash);
	signal(SIGUSR1, signal_save);

	replicate_function = default_replicate_function;
	g_hook_list_init(&data_hook, sizeof(GHook));

	loop = g_main_new(FALSE);

#ifdef HAVE_POPT_H
	pc = poptGetContext(argv[0], argc, argv, options, 0);

	while((c = poptGetNextOpt(pc)) >= 0) {
		switch(c) {
		case 'd': 
			{
				const char *fname = poptGetOptArg(pc);
				if(!strcmp(fname, "-")) { debugfd = stdout; break; }
				debugfd = fopen(fname, "w+"); 
			}
			break;
		case 'v':
			printf("ctrlproxy %s\n", PACKAGE_VERSION);
			printf("(c) 2002-2003 Jelmer Vernooij <jelmer@nl.linux.org>\n");
			return 0;
		}
	}
#endif

	if(gethostname(my_hostname, MAXHOSTNAMELEN) != 0) {
		g_log(G_LOG_DOMAIN, G_LOG_LEVEL_ERROR, "Can't figure out hostname of local host!\n");
		return 1;
	}

	if(logfile) {
		f_logfile = fopen(logfile, "a+");
		if(!f_logfile) g_warning("Can't open logfile %s!", logfile);
	}

	add_log_domain("GLib");
	add_log_domain("ctrlproxy");

	g_log(G_LOG_DOMAIN, G_LOG_LEVEL_INFO, "Logfile opened");

	if(daemon) {

#ifdef SIGTTOU
		signal(SIGTTOU, SIG_IGN);
#endif

#ifdef SIGTTIN
		signal(SIGTTIN, SIG_IGN);
#endif

#ifdef SIGTSTP
		signal(SIGTSTP, SIG_IGN);
#endif

		if(fork() != 0)exit(0);

		setsid();
		signal(SIGHUP, SIG_IGN);
		close(STDIN_FILENO);
		close(STDOUT_FILENO);
		close(STDERR_FILENO);
		daemon = 1;
	} else if(!f_logfile)f_logfile = stdout;

	if(rcfile) configuration_file = strdup(rcfile);
	else { 
		asprintf(&configuration_file, "%s/.ctrlproxyrc", g_get_home_dir());
	}

	readConfig(configuration_file);

	if(!g_module_supported()) {
		g_warning("DSO's not supported on this platform. Not loading any modules");
	} else {
		cur = xmlNode_plugins->xmlChildrenNode;
		while(cur) {

			if(xmlIsBlankNode(cur) || !strcmp(cur->name, "comment")){ cur = cur->next; continue; }

			g_assert(!strcmp(cur->name, "plugin"));

			if(xmlHasProp(cur, "autoload"))load_plugin(cur);

			cur = cur->next;
		}
	}

	cur = xmlNode_networks->xmlChildrenNode;
	while(cur) {
		char *autoconnect;
		if(xmlIsBlankNode(cur) || !strcmp(cur->name, "comment")){ cur = cur->next; continue; }
		g_assert(!strcmp(cur->name, "network"));

		autoconnect = xmlGetProp(cur, "autoconnect");
		if(autoconnect && !strcmp(autoconnect, "1"))connect_to_server(cur);
		xmlFree(autoconnect);

		cur = cur->next;
	}


	g_timeout_add(1000 * 300, ping_loop, NULL);
	g_main_run(loop);

	return 0;
}


